@@ -1,3 +1,6 @@
+# Note: these tests are part of the self test, do not use or import pytest functionality here.
+#       See borg.selftest for details. If you add/remove test methods, update SELFTEST_COUNT
+
 import base64
 import hashlib
 import io
@@ -10,9 +13,6 @@ from .. import hashindex
 from ..crypto.file_integrity import IntegrityCheckedFile, FileIntegrityError
 from . import BaseTestCase
 
-# Note: these tests are part of the self test, do not use or import py.test functionality here.
-#       See borg.selftest for details. If you add/remove test methods, update SELFTEST_COUNT
-
 
 def H(x):
     # make some 32byte long thing that depends on x
@@ -71,6 +71,22 @@ class HashIndexTestCase(BaseTestCase):
         del idx
         self.assert_equal(len(cls.read(idx_name.name)), 0)
 
+        idx = cls()
+        # Test setdefault - set non-existing key
+        idx.setdefault(H(0), make_value(42))
+        assert H(0) in idx
+        assert idx[H(0)] == make_value(42)
+        # Test setdefault - do not set existing key
+        idx.setdefault(H(0), make_value(23))
+        assert H(0) in idx
+        assert idx[H(0)] == make_value(42)
+        # Test setdefault - get-like return value, key not present
+        assert idx.setdefault(H(1), make_value(23)) == make_value(23)
+        # Test setdefault - get-like return value, key present
+        assert idx.setdefault(H(0), make_value(23)) == make_value(42)
+        # clean up setdefault test
+        del idx
+
     def test_nsindex(self):
         self._generic_test(NSIndex, lambda x: (x, x),
                            '85f72b036c692c8266e4f51ccf0cff2147204282b5e316ae508d30a448d88fef')
@@ -89,7 +105,7 @@ class HashIndexTestCase(BaseTestCase):
         for x in range(n):
             idx[H(x)] = x, x
         idx.write(idx_name.name)
-        self.assert_true(initial_size < os.path.getsize(idx_name.name))
+        assert initial_size < os.path.getsize(idx_name.name)
         for x in range(n):
             del idx[H(x)]
         self.assert_equal(len(idx), 0)
@@ -192,7 +208,7 @@ class HashIndexRefcountingTestCase(BaseTestCase):
         idx = ChunkIndex()
         idx[H(1)] = ChunkIndex.MAX_VALUE - 1, 1, 2
 
-        # 5 is arbitray, any number of incref/decrefs shouldn't move it once it's limited
+        # 5 is arbitrary, any number of incref/decrefs shouldn't move it once it's limited
         for i in range(5):
             # first incref to move it to the limit
             refcount, *_ = idx.incref(H(1))
@@ -321,7 +337,8 @@ class HashIndexDataTestCase(BaseTestCase):
         idx1[H(2)] = 2**31 - 1, 0, 0
         idx1[H(3)] = 4294962296, 0, 0  # 4294962296 is -5000 interpreted as an uint32_t
 
-        assert self._serialize_hashindex(idx1) == self.HASHINDEX
+        serialized = self._serialize_hashindex(idx1)
+        assert self._unpack(serialized) == self._unpack(self.HASHINDEX)
 
     def test_read_known_good(self):
         idx1 = self._deserialize_hashindex(self.HASHINDEX)
@@ -509,3 +526,45 @@ class AllIndexTestCase(BaseTestCase):
     def test_max_load_factor(self):
         assert NSIndex.MAX_LOAD_FACTOR < 1.0
         assert ChunkIndex.MAX_LOAD_FACTOR < 1.0
+
+
+class IndexCorruptionTestCase(BaseTestCase):
+    def test_bug_4829(self):
+
+        from struct import pack
+
+        def HH(x, y):
+            # make some 32byte long thing that depends on x and y.
+            # same x will mean a collision in the hashtable as bucket index is computed from
+            # first 4 bytes. giving a specific x targets bucket index x.
+            # y is to create different keys and does not go into the bucket index calculation.
+            # so, same x + different y --> collision
+            return pack('<IIQQQ', x, y, 0, 0, 0)  # 2 * 4 + 3 * 8 == 32
+
+        idx = NSIndex()
+
+        # create lots of colliding entries
+        for y in range(700):  # stay below max load to not trigger resize
+            idx[HH(0, y)] = (0, y)
+
+        assert idx.size() == 1031 * 40 + 18  # 1031 buckets + header
+
+        # delete lots of the collisions, creating lots of tombstones
+        for y in range(400):  # stay above min load to not trigger resize
+            del idx[HH(0, y)]
+
+        # create lots of colliding entries, within the not yet used part of the hashtable
+        for y in range(330):  # stay below max load to not trigger resize
+            # at y == 259 a resize will happen due to going beyond max EFFECTIVE load
+            # if the bug is present, that element will be inserted at the wrong place.
+            # and because it will be at the wrong place, it can not be found again.
+            idx[HH(600, y)] = 600, y
+
+        # now check if hashtable contents is as expected:
+
+        assert [idx.get(HH(0, y)) for y in range(400, 700)] == [(0, y) for y in range(400, 700)]
+
+        assert [HH(0, y) in idx for y in range(400)] == [False for y in range(400)]  # deleted entries
+
+        # this will fail at HH(600, 259) if the bug is present.
+        assert [idx.get(HH(600, y)) for y in range(330)] == [(600, y) for y in range(330)]
