@@ -37,13 +37,13 @@ from .helpers import bin_to_hex
 from .helpers import safe_ns
 from .helpers import ellipsis_truncate, ProgressIndicatorPercent, log_multi
 from .helpers import msgpack
+from .helpers import workarounds
 from .patterns import PathPrefixPattern, FnmatchPattern, IECommand
 from .item import Item, ArchiveItem
 from .platform import acl_get, acl_set, set_flags, get_flags, swidth, hostname
 from .remote import cache_if_remote
 from .repository import Repository, LIST_SCAN_LIMIT
 
-has_lchmod = hasattr(os, 'lchmod')
 has_link = hasattr(os, 'link')
 
 flags_normal = os.O_RDONLY | getattr(os, 'O_BINARY', 0)
@@ -97,11 +97,15 @@ class Statistics:
         if dt is None or now - self.last_progress > dt:
             self.last_progress = now
             if self.output_json:
-                data = self.as_dict()
+                if not final:
+                    data = self.as_dict()
+                    data['path'] = remove_surrogates(item.path if item else '')
+                else:
+                    data = {}
                 data.update({
                     'time': time.time(),
                     'type': 'archive_progress',
-                    'path': remove_surrogates(item.path if item else ''),
+                    'finished': final,
                 })
                 msg = json.dumps(data)
                 end = '\n'
@@ -225,19 +229,21 @@ class DownloadPipeline:
                     # corresponding hardlink slave is selected (== is extracted).
                     # due to a side effect of the filter() call, we now have hardlink_masters dict populated.
                     for item in items:
-                        if 'chunks' in item:  # regular file, maybe a hardlink master
-                            _preload(item.chunks)
-                            # if this is a hardlink master, remember that we already preloaded it:
-                            if 'source' not in item and hardlinkable(item.mode) and item.get('hardlink_master', True):
-                                masters_preloaded.add(item.path)
-                        elif 'source' in item and hardlinkable(item.mode):  # hardlink slave
-                            source = item.source
-                            if source not in masters_preloaded:
-                                # we only need to preload *once* (for the 1st selected slave)
-                                chunks, _ = hardlink_masters[source]
-                                if chunks is not None:
-                                    _preload(chunks)
-                                masters_preloaded.add(source)
+                        if hardlinkable(item.mode):
+                            source = item.get('source')
+                            if source is None:  # maybe a hardlink master
+                                if 'chunks' in item:
+                                    _preload(item.chunks)
+                                # if this is a hl master, remember that we already preloaded all chunks of it (if any):
+                                if item.get('hardlink_master', True):
+                                    masters_preloaded.add(item.path)
+                            else:  # hardlink slave
+                                if source not in masters_preloaded:
+                                    # we only need to preload *once* (for the 1st selected slave)
+                                    chunks, _ = hardlink_masters[source]
+                                    if chunks is not None:
+                                        _preload(chunks)
+                                    masters_preloaded.add(source)
                 else:
                     # easy: we do not have a filter, thus all items are selected, thus we need to preload all chunks.
                     for item in items:
@@ -735,10 +741,18 @@ Utilization of max. archive size: {csize_max:.0%}
             pass
         if fd:
             os.fchmod(fd, item.mode)
-        elif not symlink:
-            os.chmod(path, item.mode)
-        elif has_lchmod:  # Not available on Linux
-            os.lchmod(path, item.mode)
+        else:
+            # To check whether a particular function in the os module accepts False for its
+            # follow_symlinks parameter, the in operator on supports_follow_symlinks should be
+            # used. However, os.chmod is special as some platforms without a working lchmod() do
+            # have fchmodat(), which has a flag that makes it behave like lchmod(). fchmodat()
+            # is ignored when deciding whether or not os.chmod should be set in
+            # os.supports_follow_symlinks. Work around this by using try/except.
+            try:
+                os.chmod(path, item.mode, follow_symlinks=False)
+            except (NotImplementedError, SystemError):
+                if not symlink:
+                    os.chmod(path, item.mode)
         mtime = item.mtime
         if 'atime' in item:
             atime = item.atime
@@ -775,7 +789,6 @@ Utilization of max. archive size: {csize_max:.0%}
                 try:
                     xattr.setxattr(fd or path, k, v, follow_symlinks=False)
                 except OSError as e:
-                    msg_format = '%s: when setting extended attribute %s: %%s' % (path, k.decode())
                     if e.errno == errno.E2BIG:
                         err_str = 'too big for this filesystem'
                     elif e.errno == errno.ENOTSUP:
@@ -790,7 +803,7 @@ Utilization of max. archive size: {csize_max:.0%}
                         # EACCES: permission denied to set this specific xattr (this may happen related to security.* keys)
                         # EPERM: operation not permitted
                         err_str = os.strerror(e.errno)
-                    logger.warning(msg_format % err_str)
+                    logger.warning('%s: when setting extended attribute %s: %s', path, k.decode(), err_str)
                     set_ec(EXIT_WARNING)
         # bsdflags include the immutable flag and need to be set last:
         if not self.nobsdflags and 'bsdflags' in item:
@@ -1115,6 +1128,12 @@ Utilization of max. archive size: {csize_max:.0%}
                 raise
             # Was this EPERM due to the O_NOATIME flag? Try again without it:
             return os.open(path, flags_normal)
+        except OSError as exc:
+            # O_NOATIME causes EROFS when accessing a volume shadow copy in WSL1
+            if 'retry_erofs' in workarounds and exc.errno == errno.EROFS and flags_noatime != flags_normal:
+                return os.open(path, flags_normal)
+            else:
+                raise
 
 
 def valid_msgpacked_dict(d, keys_serialized):
@@ -1576,20 +1595,25 @@ class ArchiveChecker:
                 if state > 0:
                     unpacker.resync()
                 for chunk_id, cdata in zip(items, repository.get_many(items)):
-                    data = self.key.decrypt(chunk_id, cdata)
-                    unpacker.feed(data)
                     try:
+                        data = self.key.decrypt(chunk_id, cdata)
+                        unpacker.feed(data)
                         for item in unpacker:
                             valid, reason = valid_item(item)
                             if valid:
                                 yield Item(internal_dict=item)
                             else:
                                 report('Did not get expected metadata dict when unpacking item metadata (%s)' % reason, chunk_id, i)
+                    except IntegrityError as integrity_error:
+                        # key.decrypt() detected integrity issues.
+                        # maybe the repo gave us a valid cdata, but not for the chunk_id we wanted.
+                        # or the authentication of cdata failed, meaning the encrypted data was corrupted.
+                        report(str(integrity_error), chunk_id, i)
                     except RobustUnpacker.UnpackerCrashed:
                         report('Unpacker crashed while unpacking item metadata, trying to resync...', chunk_id, i)
                         unpacker.resync()
                     except Exception:
-                        report('Exception while unpacking item metadata', chunk_id, i)
+                        report('Exception while decrypting or unpacking item metadata', chunk_id, i)
                         raise
                     i += 1
 
@@ -1623,13 +1647,19 @@ class ArchiveChecker:
                 logger.info('Analyzing archive {} ({}/{})'.format(info.name, i + 1, num_archives))
                 archive_id = info.id
                 if archive_id not in self.chunks:
-                    logger.error('Archive metadata block is missing!')
+                    logger.error('Archive metadata block %s is missing!', bin_to_hex(archive_id))
                     self.error_found = True
                     del self.manifest.archives[info.name]
                     continue
                 mark_as_possibly_superseded(archive_id)
                 cdata = self.repository.get(archive_id)
-                data = self.key.decrypt(archive_id, cdata)
+                try:
+                    data = self.key.decrypt(archive_id, cdata)
+                except IntegrityError as integrity_error:
+                    logger.error('Archive metadata block %s is corrupted: %s', bin_to_hex(archive_id), integrity_error)
+                    self.error_found = True
+                    del self.manifest.archives[info.name]
+                    continue
                 archive = ArchiveItem(internal_dict=msgpack.unpackb(data))
                 if archive.version != 1:
                     raise Exception('Unknown archive metadata version')
