@@ -27,7 +27,7 @@ from .helpers import hostname_is_unique
 from .helpers import replace_placeholders
 from .helpers import sysinfo
 from .helpers import format_file_size
-from .helpers import truncate_and_unlink
+from .helpers import safe_unlink
 from .helpers import prepare_subprocess_env
 from .logger import create_logger, setup_logging
 from .helpers import msgpack
@@ -728,11 +728,11 @@ This problem will go away as soon as the server has been upgraded to 1.0.7+.
             args = unpacked.get(b'exception_args')
 
             if error == 'DoesNotExist':
-                raise Repository.DoesNotExist(self.location.orig)
+                raise Repository.DoesNotExist(self.location.processed)
             elif error == 'AlreadyExists':
-                raise Repository.AlreadyExists(self.location.orig)
+                raise Repository.AlreadyExists(self.location.processed)
             elif error == 'CheckNeeded':
-                raise Repository.CheckNeeded(self.location.orig)
+                raise Repository.CheckNeeded(self.location.processed)
             elif error == 'IntegrityError':
                 if old_server:
                     raise IntegrityError('(not available)')
@@ -752,9 +752,9 @@ This problem will go away as soon as the server has been upgraded to 1.0.7+.
                 raise Repository.ParentPathDoesNotExist(args[0].decode())
             elif error == 'ObjectNotFound':
                 if old_server:
-                    raise Repository.ObjectNotFound('(not available)', self.location.orig)
+                    raise Repository.ObjectNotFound('(not available)', self.location.processed)
                 else:
-                    raise Repository.ObjectNotFound(args[0].decode(), self.location.orig)
+                    raise Repository.ObjectNotFound(args[0].decode(), self.location.processed)
             elif error == 'InvalidRPCMethod':
                 if old_server:
                     raise InvalidRPCMethod('(not available)')
@@ -1144,7 +1144,7 @@ class RepositoryCache(RepositoryNoCache):
                 fd.write(packed)
         except OSError as os_error:
             try:
-                truncate_and_unlink(file)
+                safe_unlink(file)
             except FileNotFoundError:
                 pass  # open() could have failed as well
             if os_error.errno == errno.ENOSPC:
